@@ -9,12 +9,26 @@ from enum import Enum
 import json
 import logging
 from copy import deepcopy
+import pandas as pd
+import threading
 
 VERSION = "2021-08-16"
 
 global notion_instance
 
 
+def ensure_page_id(page_id):
+    if "http" in page_id:
+        page_id = page_id.split("/")[-1].split("-")[-1]
+    return page_id
+
+
+def ensure_database_id(database_id):
+    if "http" in database_id:
+        database_id = database_id.split("/")[-1].split("?")[0]
+    return database_id
+
+
 class Notion(InDriver, OutDriver):
     __client = None
 
@@ -84,13 +98,8 @@ class Notion(InDriver, OutDriver):
         def __from_dict(self, data):
             return from_dict(data_class=Database, data=data)
 
-        def __ensure_database_id(self, database_id):
-            if "http" in database_id:
-                database_id = database_id.split("/")[-1].split("?")[0]
-            return database_id
-
         def query(self, database_id, query={}):
-            database_id = self.__ensure_database_id(database_id)
+            database_id = ensure_database_id(database_id)
             ret = []
             results = self.client.databases.query(database_id=database_id, **query).get(
                 "results"
@@ -113,7 +122,7 @@ class Notion(InDriver, OutDriver):
             return self.retrieve(database_id)
 
         def retrieve(self, database_id):
-            database_id = self.__ensure_database_id(database_id)
+            database_id = ensure_database_id(database_id)
             raw = self.client.databases.retrieve(database_id)
             return self.__from_dict(raw)
 
@@ -121,17 +130,18 @@ class Notion(InDriver, OutDriver):
         def __from_dict(self, data):
             return from_dict(data_class=Page, data=data)
 
-        def __ensure_page_id(self, page_id):
-            if "http" in page_id:
-                page_id = page_id.split("/")[-1].split("-")[-1]
-            return page_id
-
-        def create(self, page):
+        def create_from_page(self, page):
             payload = self.parent.to_dict(page)
             return self.__from_dict(self.client.pages.create(**payload))
 
+        def create(self, database_id, title):
+            database_id = ensure_database_id(database_id)
+            new_page = Page.new(database_id=database_id)
+            new_page.title("Name", title)
+            return new_page.create()
+
         def retrieve(self, page_id: str):
-            page_id = self.__ensure_page_id(page_id)
+            page_id = ensure_page_id(page_id)
             page = self.client.pages.retrieve(page_id=page_id)
             return self.__from_dict(page)
 
@@ -173,7 +183,8 @@ class Notion(InDriver, OutDriver):
 
         def children(self, block_id: str):
             ret = []
-            for r in self.client.blocks.children.list(block_id=block_id).get("results"):
+            blocks = self.client.blocks.children.list(block_id=block_id).get("results")
+            for r in blocks:
                 ret.append(self.__from_dict(r))
             return ret
 
@@ -383,6 +394,12 @@ class Equation(__BaseDataClass):
         return from_dict(data_class=cls, expression=expression)
 
 
+@dataclass
+class UserMention(__BaseDataClass):
+    type: str
+    user: User
+
+
 @dataclass
 class RichText(__BaseDataClass):
     plain_text: str
@@ -390,7 +407,7 @@ class RichText(__BaseDataClass):
     annotations: Annotation = field(default_factory=Annotation)
     type: str = "text"
     text: Optional[Text] = None
-    user: Optional[User] = None
+    mention: Optional[UserMention] = None
     page: Optional[object] = None
     database: Optional[object] = None
     date: Optional[Date] = None
@@ -645,7 +662,7 @@ class Database(__BaseDataClass):
     title: List[RichText]
     properties: object
     parent: Parent
-    icon: Optional[Union[File, Emoji]] = None
+    # icon: Optional[Emoji] = None  # TODO: Handle File
     cover: Optional[File] = None
     created_time: str = None
     last_edited_time: str = None
@@ -664,6 +681,21 @@ class Database(__BaseDataClass):
         for k in self.properties:
             self.properties[k] = DatabasePropertyFactory.new(self.properties[k])
 
+    def __repr__(self):
+        try:
+            display(self.df())  # noqa: F821
+            return ""
+        except:  # noqa: E722
+            return str(self)
+
+    def schema(self):
+        copied = deepcopy(self.properties)
+        d = []
+        for i in copied:
+            p = copied[i]
+            d.append({"Name": i, "Type": p.type})
+        return pd.DataFrame(self.notion.to_dict(d))
+
     @classmethod
     def new(cls, title: str, page_id: str = None):
         parent = None
@@ -672,23 +704,44 @@ class Database(__BaseDataClass):
         else:
             parent = Parent.new_workspace_parent()
 
-        return cls(
-            title=[RichText.new_text(content=title)], parent=parent, properties={}
-        )
+        new_db = cls(title=[RichText.new(title)], parent=parent, properties={})
+        new_db.add_property("Name", "title")
+        return new_db
 
     @classmethod
     def from_dict(cls, data):
         return from_dict(data_class=cls, data=data)
 
-    def set_emoji_icon(
-        self, data
-    ):  # TODO: Fix, seems like there is an issue with notion-client / httpx maybe.
-        self.icon = Emoji(data)
+    # def set_emoji_icon(
+    #    self, data
+    # ):  # TODO: Fix, seems like there is an issue with notion-client / httpx maybe.
+    #    self.icon = Emoji(data)
+
+    def duplicate(self):
+        new_db = self.create()
+        pages = self.query()
+        for page in pages:
+            page.parent.database_id = new_db.id
+            page.duplicate()
+
+        return new_db
 
     def add_property(self, col_name, type_name):
         data = {"type": type_name, "name": col_name, "id": ""}
         self.properties[col_name] = DatabasePropertyFactory.new(data)
 
+    def df(self):
+        pages = self.query()
+        frames = []
+
+        for page in pages:
+            frame = {}
+            for prop in page.properties:
+                frame[prop] = str(page.properties[prop])
+            frames.append(frame)
+
+        return pd.DataFrame(frames)
+
     def query(self, query={}):
         return self.notion.databases.query(self.id, query)
 
@@ -721,6 +774,9 @@ class PagePropertyTitle(PageProperty):
     def set_text(self, content: str):
         self.title = [RichText.new_text(content)]
 
+    def __str__(self):
+        return " ".join([v.plain_text for v in self.title])
+
 
 @dataclass
 class PagePropertyText(PageProperty):
@@ -731,6 +787,9 @@ class PagePropertyText(PageProperty):
     def new(cls, content: str) -> "PagePropertyText":
         return cls(rich_text=[RichText.new(content)])
 
+    def __str__(self):
+        return " ".join([v.plain_text for v in self.rich_text])
+
 
 @dataclass
 class PagePropertyNumber(PageProperty):
@@ -741,6 +800,9 @@ class PagePropertyNumber(PageProperty):
     def new(cls, number: Number) -> "PagePropertyNumber":
         return cls(number=number)
 
+    def __str__(self):
+        return str(self.number)
+
 
 @dataclass
 class PagePropertySelect_Value:
@@ -764,6 +826,9 @@ class PagePropertySelect(PageProperty):
     def new(cls, select: str, color: str = "default") -> "PagePropertySelect":
         return cls(select=PagePropertySelect_Value.new(name=select, color=color))
 
+    def __str__(self):
+        return str(self.select.name) if self.select else ""
+
 
 @dataclass
 class PagePropertyMultiSelect(PageProperty):
@@ -780,6 +845,9 @@ class PagePropertyMultiSelect(PageProperty):
             multi_select=[PagePropertySelect_Value.new(v, color) for v in values]
         )
 
+    def __str__(self):
+        return ", ".join([v.name for v in self.multi_select])
+
 
 @dataclass
 class PagePropertyDate_Value:
@@ -802,6 +870,9 @@ class PagePropertyDate(PageProperty):
     def new(cls, date: str) -> "PagePropertyDate":
         return cls(date=PagePropertyDate_Value.new(date))
 
+    def __str__(self):
+        return str(self.date.start) if self.date else ""
+
 
 @dataclass
 class PagePropertyFormula(PageProperty):
@@ -815,6 +886,8 @@ class PagePropertyFormula(PageProperty):
 
     # TODO: classmethod new
 
+    # TODO: classmethod new
+
 
 @dataclass
 class PagePropertyRelation(PageProperty):
@@ -845,6 +918,9 @@ class PagePropertyPeople(PageProperty):
     def new(cls, people: List[str]) -> "PagePropertyPeople":
         return cls(people=[User.new(v) for v in people])
 
+    def __str__(self):
+        return " ".join([v.name for v in self.people])
+
 
 @dataclass
 class PagePropertyFiles(PageProperty):
@@ -855,6 +931,9 @@ class PagePropertyFiles(PageProperty):
     def new(cls, files: List[str]) -> "PagePropertyFiles":
         return cls(files=[File.new(v) for v in files])
 
+    def __str__(self):
+        return ", ".join([v.name for v in self.files])
+
 
 @dataclass
 class PagePropertyCheckbox(PageProperty):
@@ -865,6 +944,9 @@ class PagePropertyCheckbox(PageProperty):
     def new(cls, checkbox: bool) -> "PagePropertyCheckbox":
         return cls(checkbox=checkbox)
 
+    def __str__(self):
+        return str(self.checkbox)
+
 
 @dataclass
 class PagePropertyUrl(PageProperty):
@@ -875,6 +957,9 @@ class PagePropertyUrl(PageProperty):
     def new(cls, url: str) -> "PagePropertyUrl":
         return cls(url=url)
 
+    def __str__(self):
+        return str(self.url)
+
 
 @dataclass
 class PagePropertyEmail(PageProperty):
@@ -885,6 +970,9 @@ class PagePropertyEmail(PageProperty):
     def new(cls, email: str) -> "PagePropertyEmail":
         return cls(email=email)
 
+    def __str__(self):
+        return str(self.email)
+
 
 @dataclass
 class PagePropertyPhoneNumber(PageProperty):
@@ -895,30 +983,45 @@ class PagePropertyPhoneNumber(PageProperty):
     def new(cls, phone_number: str) -> "PagePropertyPhoneNumber":
         return cls(phone_number=phone_number)
 
+    def __str__(self):
+        return str(self.phone_number)
+
 
 @dataclass
 class PagePropertyCreatedBy(PageProperty):
     created_by: User = field(default_factory=User)
     type: str = "created_by"
 
+    def __str__(self):
+        return str(self.created_by.name) if self.created_by else ""
+
 
 @dataclass
 class PagePropertyCreatedTime(PageProperty):
     created_time: str = None
     type: str = "created_time"
 
+    def __str__(self):
+        return str(self.created_time)
+
 
 @dataclass
 class PagePropertyLastEditedTime(PageProperty):
     last_edited_time: str = None
     type: str = "last_edited_time"
 
+    def __str__(self):
+        return str(self.last_edited_time)
+
 
 @dataclass
 class PagePropertyLastEditedBy(PageProperty):
     last_edited_by: User = field(default_factory=User)
     type: str = "last_edited_by"
 
+    def __str__(self):
+        return str(self.last_edited_by.name) if self.last_edited_by else ""
+
 
 class PagePropertyFactory:
     """
@@ -971,7 +1074,7 @@ class Page(__BaseDataClass):
     parent: Parent
     blocks: Optional[List["Block"]] = field(default_factory=list)
     archived: bool = False
-    icon: Optional[Union[File, Emoji]] = None
+    # icon: Optional[Emoji] = None  # TODO: Handle File
     cover: Optional[File] = None
     created_time: str = None
     last_edited_time: str = None
@@ -990,10 +1093,60 @@ class Page(__BaseDataClass):
         for k in self.properties:
             self.properties[k] = PagePropertyFactory.new(self.properties[k])
 
+    def __repr__(self):
+        try:
+            display(self.df())  # noqa: F821
+            return ""
+        except:  # noqa: E722
+            return str(self)
+
     def get_blocks(self):
         self.blocks = self.notion.blocks.children(self.id)
         return self.blocks
 
+    def duplicate(self):
+        page = deepcopy(self)
+        page.id = None
+        page.parent.type = None
+
+        to_delete = []
+        for col_name in page.properties:
+            prop = page.properties[col_name]
+            if prop.type in [
+                "last_edited_time",
+                "created_time",
+                "created_by",
+                "last_edited_by",
+            ]:
+                to_delete.append(col_name)
+            elif getattr(prop, prop.type) is None:
+                to_delete.append(col_name)
+
+        for col_name in to_delete:
+            del page.properties[col_name]
+
+        block_tree = BlockTree(self.notion.blocks.get(self.id))
+        new_page = page.create()
+        block_tree.duplicate(new_page.id)
+
+        return new_page
+
+    def df(self, pivot=False):
+        copied = deepcopy(self.properties)
+        if pivot is False:
+            d = []
+            for i in copied:
+                p = copied[i]
+
+                d.append({"Name": i, "Type": p.type, "Value": str(p)})
+            return pd.DataFrame(self.notion.to_dict(d))
+        else:
+            frame = {}
+            for i in copied:
+                p = copied[i]
+                frame[i] = str(p)
+            return pd.DataFrame(self.notion.to_dict([frame]))
+
     @classmethod
     def new(cls, page_id: str = None, database_id: str = None):
         parent = None
@@ -1028,7 +1181,7 @@ class Page(__BaseDataClass):
         return ret
 
     def create(self):
-        return self.notion.pages.create(self)
+        return self.notion.pages.create_from_page(self)
 
     """
     Properties setters
@@ -1150,6 +1303,9 @@ class Page(__BaseDataClass):
     def table_of_contents(self, *k):
         return self.__generic_block_setter("table_of_contents", *k)
 
+    def child_database(self, *k):
+        return self.__generic_block_setter("child_database", *k)
+
 
 """
 Block and Blocks properties.
@@ -1194,6 +1350,10 @@ class Block(__BaseDataClass):
             return from_dict(data_class=cls, data={"type": type, type: prop})
         return from_dict(data_class=cls, data={"type": type, type: {}})
 
+    # def __post_init__(self):
+    #     if self.has_children is True:
+    #         self.children = self.notion.blocks.children(block_id=self.id)
+
     # TODO: To delete
     # def __post_init__(self):
     #   setattr(self, self.type, BlockTypeFactory.new(self))
@@ -1329,6 +1489,10 @@ class BlockChildPage(__BaseDataClass):
 class BlockChildDatabase(__BaseDataClass):
     title: Optional[str] = ""
 
+    @classmethod
+    def new(cls, title: str) -> "BlockChildDatabase":
+        return cls(title=title)
+
 
 @dataclass
 class BlockEmbed(__BaseDataClass):
@@ -1515,3 +1679,54 @@ def del_none(d):
         for e in d:
             del_none(e)
     return d  # For convenience
+
+
+class BlockTree:
+    block: Block = None
+    children: List["BlockTree"] = None
+
+    def __init__(self, block: Block):
+        self.block = block
+        if self.block.has_children is True:
+            self.children = []
+            blocks = Notion.instance().blocks.children(self.block.id)
+            for b in blocks:
+                self.children.append(BlockTree(b))
+
+    def copy(self):
+        copied_block = deepcopy(self.block)
+        copied_block.id = None
+        copied_block.created_time = None
+        copied_block.last_edited_time = None
+        return copied_block
+
+    def duplicate_children(self, parent_id):
+        if self.children is not None:
+            to_append = []
+            for child in self.children:
+                to_append.append(child.copy())
+            new_blocks = Notion.instance().block.append(parent_id, to_append)
+            count = 0
+            for child in self.children:
+                child.duplicate_children(new_blocks[count].id)
+                count = count + 1
+
+    def duplicate(self, parent_id):
+        if self.children is not None:
+            to_append = []
+            for child in self.children:
+                to_append.append(child.copy())
+            new_blocks = Notion.instance().block.append(parent_id, to_append)
+
+            threads = []
+            count = 0
+            for child in self.children:
+                x = threading.Thread(
+                    target=child.duplicate_children, args=(new_blocks[count].id,)
+                )
+                threads.append(x)
+                x.start()
+                count = count + 1
+
+            for t in threads:
+                t.join()
